#   qobra is a simple terminal music player
#   Contribute at agoussas@espol.edu.ec

#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''This module contains functions to handle the program's options and events

Functions:
get_opts -- manages program options
handle_input -- handles key events
get_chunks -- yields n sized list chunks
'''

import os
import sys
import curses

sys.path.insert(1, os.path.expanduser('~/.config/qobra/'))

try:
    from config import music_dir
except ImportError:
    from qobra.config import music_dir


def usage():
    usage = '''\
Usage

qobra [-d music dir] [options]

-d  music directory
-s  start in shuffle mode
-h  see this message'''
    sys.exit(usage)


def get_opts():
    '''Manages program options

    Returns the path to the music directory along with any set
    options, or exits with an error message if the program was called
    with no arguments and directory is not defined in
    ~/.config/qobra/config.py or an invalid number of options was
    provided.
    '''
    directory, shuffle = music_dir, False
    try:
        for i, arg in enumerate(sys.argv[1:]):
            if arg == '-h':
                usage()
            elif arg == '-d':
                directory = sys.argv[i + 2]
            elif arg == '-s':
                shuffle = True
            else:
                pass
    except IndexError:
        sys.exit('Invalid number of options')

    if not directory:
        sys.exit('Argument missing: directory\nqobra -h to see usage')

    return directory, shuffle


def handle_input(stdscr, player, searchbox, window):
    '''Handle key events

    Parameters:
    stdscr -- the screen object
    player -- the player object
    '''
    curses.halfdelay(1)
    c = stdscr.getch()

    if c == ord('q'):
        player.song.stop()
        sys.exit(0)
    elif c == curses.KEY_RESIZE:
        window.resize(player, searchbox)
    elif c == ord('l'):
        player.play()
    elif c == ord('p'):
        player.pause()
    elif c == ord('k'):
        player.move_up()
    elif c == ord('i'):
        player.move_up()
        player.play()
    elif c == ord('j'):
        player.move_down()
    elif c == ord('o'):
        player.move_down()
        player.play()
    elif c == ord('t'):
        player.toggle_mode()
    elif c == ord('/'):
        searchbox.draw_search_box()


def get_chunks(it, n):
    '''Yields n sized chunks of it

    Parameters:
    list -- list object to be divided in chunks
    n -- integer with the number of chunks to create
    '''
    for i in range(0, len(it), n):
        yield it[i:i + n]



#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from subprocess import Popen, DEVNULL, STDOUT, PIPE


class Song:
    def __init__(self, path):
        self.path = path
        self.playing = False
        self.process = None

    @property
    def name(self):
        '''Sets the song name.'''
        return os.path.basename(self.path)

    @property
    def running(self):
        '''Returns True if current process has finished.'''
        return self.process.poll() is None

    def play(self):
        '''Spawns a new mpg123 process to play a song.'''
        self.playing = True
        self.process = Popen(
            ['mpg123', self.path, '--title', '--name', self.name],
            stdin=PIPE,
            stdout=DEVNULL,
            stderr=STDOUT)

    def pause(self):
        '''Sends space key to current process to stop song.

        mpg123 also accepts "s" to toggle playing state of song.
        '''
        supported_terminals = ['xterm', 'rxvt', 'screen', 'iris-ansi']
        if os.environ['TERMINAL'] in supported_terminals or os.environ[
                'TERM'] in supported_terminals:
            # TODO test this
            xdotool = Popen(['xdotool', 'search', '--name', self.name],
                            stdout=PIPE)
            winid, _ = xdotool.communicate()
            pauser = Popen(['xdotool', 'key', '--window', winid, 'space'])
            pauser.wait()
            self.playing = not self.playing
        else:
            self.stop()

    def stop(self):
        '''Sends SIGTERM signal to current process.'''
        try:
            self.playing = False
            self.process.terminate()
        except AttributeError:
            pass



#  @@@@@@  @@@@@@ @@@@@@@ @@@@@@@  @@@@@@
# @@!  @@@@@!  @@@@@!  @@@@@!  @@@@@!  @@@
# @!@  !@!@!@  !@!@!@!@!@ @!@!!@! @!@!@!@!
# !!:!!:!:!!:  !!!!!:  !!!!!: :!! !!:  !!!
#  : :. :::: :. : :: : ::  :   : : :   : :
#
# This is the configuration file for qobra.
# qobra will look for a configuration file in ~/.config/qobra
# and load the default configuration if it is not found.

# Music directory
# Set this to avoid having to type the path in the command
# line
music_dir = ''

# Colors
# Options are:
# black, red, green, yellow, blue, magenta, cyan and white

# Window colors (the border color is determined by window_fg)
window_fg = 'red'
window_bg = 'black'

# Songs colors
songs_fg = 'red'
songs_bg = 'black'

# Selected song colors
current_song_fg = 'black'
current_song_bg = 'red'

##############
# Statusline #
##############

# Current status indicator
# Playing  colors
playing_fg = 'green'
playing_bg = 'black'
# Paused colors
paused_fg = 'red'
paused_bg = 'black'

# Current song colors
statusbar_song_fg = 'white'
statusbar_song_bg = 'black'

# Mode indicator
# Normal mode colors
normal_mode_fg = 'cyan'
normal_mode_bg = 'black'
# Suffle mode colors
shuffle_mode_fg = 'yellow'
shuffle_mode_bg = 'black'

# Search bar

# Search placeholder colors
search_fg = 'magenta'
search_bg = 'black'

# Search box colors
search_box_fg = 'red'
search_box_bg = 'black'

# MISC
# Draw a border around the window?
border = True

# Background fill character
# This can't be an empty string
bg_char = ' '



#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import curses
from curses.textpad import Textbox

sys.path.insert(1, os.path.expanduser('~/.config/qobra/'))

try:
    from config import search_fg, search_bg, search_box_fg, search_box_bg
except ImportError:
    from qobra.config import (search_fg, search_bg, search_box_fg,
                              search_box_bg)


class SearchBox:
    '''Text box to filter songs.

    Creates a text box at the bottom of the window when the
    search event is triggered.

    Attributes:
    stdscr -- the screen object
    player -- the player object
    colors -- mapping of curses colors
    height -- the TextBox instance height
    widht -- the TextBox instance width
    x -- the TextBox instance x offset
    y -- the TextBox instance y offset
    action -- the action performed when exiting the textbox
    '''
    def __init__(self, player, stdscr, colors):
        self.stdscr = stdscr
        self.colors = colors
        self.player = player

        self.height = 1
        self.width = self.stdscr.getmaxyx()[1] - 11
        self.x, self.y = 10, self.stdscr.getmaxyx()[0] - 2
        self.searched = False

    def draw_search_box(self):
        '''Draws the search box.

        Draws a search box at the bottom of the screen and let's the user
        filter songs. The contents of the search box are then dispatched
        to a filter function to update the current song and chunk.
        '''
        curses.init_pair(9, self.colors[search_fg], self.colors[search_bg])
        curses.init_pair(10, self.colors[search_box_fg],
                         self.colors[search_box_bg])
        self.stdscr.addstr(self.y, 2, 'Search: ', curses.color_pair(9))
        self.stdscr.refresh()

        editwin = curses.newwin(self.height, self.width, self.y, self.x)
        editwin.bkgd(' ', curses.color_pair(10))

        box = Textbox(editwin)
        box.edit(self.__handle_keys)

        if self.searched:
            self.__filter_songs(box.gather())

        self.searched = False

    def __filter_songs(self, match):
        '''Updates the current song and chunk.

        Uses the string provided in the match parameter to update the
        current chunk to the one containing the song name that matches
        it.
        '''
        for song in self.player.songs:
            if match.strip().lower() in song.strip().lower():
                matched_song = song
                for idx, group in enumerate(self.player.groups):
                    if matched_song in group:
                        self.player.group_index = idx
                        self.player.current_group = self.player.groups[
                            self.player.group_index]
                        self.player.song_index = self.player.current_group.index(
                            matched_song)
                        break
                break

    def __handle_keys(self, ch):
        '''Helper function for the edit() method

        Handles key events that occur when the search box is spawned
        '''
        if ch == 27:  # Escape key
            return 7
        elif ch == 10:  # Enter key
            self.searched = True
            return 7
        else:
            return ch



#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import curses
import random

import qobra.aidees as f
from qobra.song import Song

sys.path.insert(1, os.path.expanduser('~/.config/qobra/'))

try:
    from config import (
        shuffle_mode_fg,
        shuffle_mode_bg,
        normal_mode_fg,
        normal_mode_bg,
        statusbar_song_fg,
        statusbar_song_bg,
        paused_fg,
        paused_bg,
        playing_fg,
        playing_bg,
        songs_fg,
        songs_bg,
        current_song_fg,
        current_song_bg,
    )
except ImportError:
    try:
        from qobra.config import (
            shuffle_mode_fg,
            shuffle_mode_bg,
            normal_mode_fg,
            normal_mode_bg,
            statusbar_song_fg,
            statusbar_song_bg,
            paused_fg,
            paused_bg,
            playing_fg,
            playing_bg,
            songs_fg,
            songs_bg,
            current_song_fg,
            current_song_bg,
        )
    except ImportError:
        raise ImportError('Couldn\'t find path to config file')


class Player:
    '''Manages state of the program.

    Attributes:
    root -- the path to the user's music directory
    song_index -- index of the curent song
    group_index -- index of the current group
    songs -- list with the songs in the provided path
    groups -- sublists of songs divided in window-sized chunks
    current_group -- current group displayed on the screen
    song -- name of the currently playing song
    '''
    def __init__(self, root, shuffle, stdscr, colors, window):
        self.stdscr = stdscr
        self.colors = colors
        self.shuffle = shuffle
        self.window = window

        self.songs = [os.path.join(root, song) for song in os.listdir(root)]
        self.song_index, self.group_index = 0, 0
        self.groups = list(f.get_chunks(self.songs, self.window.height - 3))
        self.current_group = self.groups[self.group_index]
        self.song = Song(self.current_group[self.song_index])
        self.init_colors()

    def display_songs(self):
        '''Displays songs on the screen.

        Display the songs contained in the current group. It
        starts drawing them 1 line after the window start to
        compensate for the border. The curren song is given
        different colors.
        '''
        x_offset, y_offset = 3, 1
        for idx, song in enumerate(self.current_group):
            song = os.path.basename(song)
            free_space = self.window.width - len(song) - x_offset
            song_color = curses.color_pair(
                1) if idx == self.song_index else curses.color_pair(2)
            self.stdscr.addnstr(idx + y_offset, x_offset,
                                song + ' ' * free_space, self.window.width - 5,
                                song_color)

    def update_group_size(self, new_height):
        '''Updates the chunk size on window resize

        Parameters:
        new_height -- the new height of the window
        '''
        self.groups = list(f.get_chunks(self.songs, new_height - 3))

    def update(self):
        '''Play next song after song_index one ends

        Checks if a song is playing and if the last mpg123 process has
        terminated. If a song is playing and there are no living processes,
        play the next song.
        '''
        if self.song.playing and not self.song.running:
            self.song_index = self.playing_song_index
            self.move_down()
            self.play()

    def play(self):
        '''
        Plays a song. Stop the current process first to make sure no duplicate
        processes are running.
        '''
        self.song.stop()
        if self.shuffle:
            this_song = self.select_random()
        else:
            this_song = self.current_group[self.song_index]
        self.playing_song_index = self.song_index
        self.song = Song(this_song)
        self.song.play()

    def select_random(self):
        '''Set a random song to be the song_index one'''
        random_song = random.choice(self.songs)
        for group_number, group in enumerate(self.groups):
            for song_number, song in enumerate(group):
                if song == random_song:
                    self.group_index = group_number
                    self.current_group = self.groups[self.group_index]
                    self.song_index = song_number

        return random_song

    def toggle_mode(self):
        '''Toggles between Normal and Shuffle mode.'''
        self.shuffle = not self.shuffle

    def pause(self):
        '''Pauses the player.'''
        # im working on a way to send keys to mpg123 portably
        self.song.pause()

    def move_down(self):
        '''Moves down a song.

        If the song is the last one in the current group,
        moves to the first song in the next chunk, else
        it just moves down a song.
        '''
        if self.song_index < len(self.current_group) - 1:
            self.song_index += 1
        elif self.group_index < len(self.groups) - 1:
            self.song_index = 0
            self.group_index += 1
            self.current_group = self.groups[self.group_index]
        else:
            self.song_index, self.group_index = 0, 0

    def move_up(self):
        '''Moves up a song

        If the song is the the first one in the current group,
        move to the last song in the previous one,
        else just move back a song.
        '''
        if self.song_index > 0:
            self.song_index -= 1
        elif self.group_index > 0:
            self.group_index -= 1
            self.current_group = self.groups[self.group_index]
        else:
            self.group_index = len(self.groups) - 1
            self.current_group = self.groups[self.group_index]
            self.song_index = len(self.current_group) - 1

    def init_colors(self):
        '''Initializes colors.'''
        # songs colors
        curses.init_pair(1, self.colors[current_song_fg],
                         self.colors[current_song_bg])
        curses.init_pair(2, self.colors[songs_fg], self.colors[songs_bg])
        # statusline
        curses.init_pair(4, self.colors[playing_fg], self.colors[playing_bg])
        curses.init_pair(5, self.colors[paused_fg], self.colors[paused_bg])
        curses.init_pair(6, self.colors[statusbar_song_fg],
                         self.colors[statusbar_song_bg])
        curses.init_pair(7, self.colors[normal_mode_fg],
                         self.colors[normal_mode_bg])
        curses.init_pair(8, self.colors[shuffle_mode_fg],
                         self.colors[shuffle_mode_bg])



#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import curses

try:
    from config import window_fg, window_bg, bg_char, border
except ImportError:
    from qobra.config import (window_fg, window_bg, bg_char, border)


class Window:
    def __init__(self, stdscr, colors):
        self.stdscr = stdscr
        self.height, self.width = stdscr.getmaxyx()
        self.colors = colors

    def draw(self, player):
        '''Draws the program window'''
        curses.init_pair(3, self.colors[window_fg], self.colors[window_bg])
        self.stdscr.bkgd(bg_char, curses.color_pair(3))
        if border:
            self.stdscr.border()
        self.draw_statusline(player)
        player.display_songs()

    def draw_statusline(self, player):
        '''Draw the statusbar at the bottom of the screen

        Display information about the song_index playing song,
        the playing status and the playing mode. It is drawn
        2 lines above the window bottom to compensate for the
        border.
        '''
        y = self.height - 2
        # player status
        status = 'Playing' if player.song.playing else 'Paused'
        status_color = 4 if player.song.playing else 5
        self.stdscr.addstr(y, 2, status, curses.color_pair(status_color))
        # name of the current song
        self.stdscr.addnstr(y, 10, player.song.name, self.width - 20,
                            curses.color_pair(6))
        # playing mode
        mode = 'Shuffle' if player.shuffle else 'Normal'
        mode_colors = 7 if not player.shuffle else 8
        self.stdscr.addstr(y, self.width - len(mode) - 1, mode,
                           curses.color_pair(mode_colors))

    def resize(self, player, searchbox):
        '''Resizes the window.

        Updates the player's current group size to fit in the new window
        height and makes sure the current song stays selected.
        '''
        self.height, self.width = self.stdscr.getmaxyx()
        player.update_group_size(self.height)
        for group in player.groups:
            if player.song in group:
                player.group_index = player.groups.index(group)
        player.current_group = player.groups[player.group_index]
        player.song_index = player.current_group.index(player.song.path)

        searchbox.width = self.stdscr.getmaxyx()[1] - 11
        searchbox.y = self.stdscr.getmaxyx()[0] - 2



#!/usr/bin/env python3

#   qobra is a simple terminal music player
#   Contribute at agoussas@espol.edu.ec

#   Copyright (C) 2020  Alexander Goussas
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import curses

import qobra.aidees as f
from qobra.player import Player
from qobra.searchbox import SearchBox
from qobra.window import Window

colors = {
    'white': curses.COLOR_WHITE,
    'black': curses.COLOR_BLACK,
    'red': curses.COLOR_RED,
    'green': curses.COLOR_GREEN,
    'yellow': curses.COLOR_YELLOW,
    'blue': curses.COLOR_BLUE,
    'magenta': curses.COLOR_MAGENTA,
    'cyan': curses.COLOR_CYAN,
}


def main(stdscr):
    '''Run the main loop.

    Parameters:
    stdscr -- the screen object used to draw the window
    '''
    curses.curs_set(False)
    path, shuffle = f.get_opts()

    window = Window(stdscr, colors)
    player = Player(path, shuffle, stdscr, colors, window)
    searchbox = SearchBox(player, stdscr, colors)

    while True:
        stdscr.erase()
        window.draw(player)
        player.update()
        stdscr.refresh()
        f.handle_input(stdscr, player, searchbox, window)


curses.wrapper(main)
if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from setuptools import setup, find_packages
import os


def get_readme(readme_file):
    with open(readme_file, 'rt', encoding='utf-8') as readme:
        contents = readme.read()

    return contents


README = os.path.join(os.path.dirname(__file__), 'README.md')

setup(name='qobra',
      version='1.1.1',
      author='Alexander Goussas',
      author_email='agoussas@espol.edu.ec',
      description='A simple music player for the command line',
      long_description=get_readme(README),
      long_description_content_type='text/markdown',
      url='https://github.com/aloussase/qobra',
      packages=find_packages(),
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
      ],
      install_requires=[],
      python_requires='>=3.6',
      entry_points={'console_scripts': ['qobra = qobra.qobra:main']})
